#version 120
#extension GL_EXT_geometry_shader4 : enable

#define PROTSTART 1.5
#define PROTEND 4.0

uniform float mol_abstractionmax;
uniform float protein_radius;
uniform float depth_cueing;
uniform float attenuate;

varying in float structural_abs[];
varying out float structural_abstraction;
varying in float illustrativenessIn[];
varying out float illustrativeness;
varying in float atom_rank[];
varying in float atom_radius[];
varying out float radius;
varying out float depth_top;
varying out float halo_radius;
varying out mat3 rotationMatrix;
varying in float atom_type[];
varying out float type;

// Pastel colors
varying in vec3 pastel_lightIn[];
varying out vec3 pastel_light;
varying in vec3 pastel_darkIn[];
varying out vec3 pastel_dark;

// Ambient Occlussion variables
uniform sampler2D ambient_occlusion0;
varying out	float ao_fact;
varying out	float ao_contrib;
#define AO_ABS_LVL0 0.0
#define AO_ABS_LVL1 0.2

// Debug variables
varying out float hint;
varying out vec3 upVector;
varying out vec3 camVector;

float mol_abstraction;
float sphere_attenuation;

varying in float atom_id[];
varying out float id;

mat4 coords = mat4(-1.0, -1.0, 0.0, 0.0, -1.0, 1.0, 0.0, 0.0, 1.0, -1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0);

void calculate_abstraction() {
	float tmp, shrink;

	/* PRE: When this program is used: 0 <= structural_abs <= 0.6 */
	if (structural_abstraction < 0.25) {
		tmp = structural_abstraction * (4.0 * 0.9);
		shrink = 1.0 - 2.0 * structural_abstraction;
		radius = shrink * (atom_radius[0] / protein_radius) * (1.0 - tmp);
		mol_abstraction = mol_abstractionmax;
	}
	else {
		radius = (atom_radius[0] / protein_radius) * 0.05;
		tmp = 1.0/(0.6-0.25) * mol_abstractionmax * (structural_abstraction-0.25);
    mol_abstraction = mol_abstractionmax - tmp;
	}
}

vec4 quaternion_inverse(in vec4 q) {
  return vec4(-q.xyz, q.w) / length(q);
}

vec4 quaternion_mult(in vec4 q1, in vec4 q2) {
  return vec4(q1.w * q2.xyz + q2.w * q1.xyz + cross(q1.xyz, q2.xyz), q1.w * q2.w - dot(q1.xyz, q2.xyz));
}
 
vec4 determine_position(in vec4 vertex, inout vec4 coord) {
  vec4 pos;
  vec4 referenceVector;
  vec4 cam;
  vec4 up;
  vec4 displ;
  vec4 camVec, camVecProj;
  vec4 proj;
	float distCam;
	float attenuation;
  vec4 s1, s1inv, s2;

  vec3 rotationAxis;
  vec3 upProj;
  float rotationAngle;
  float a, b, c, d;
  vec3 tmp;
	vec3 refVec;

  // Figure out where the camera is in model coordinates
  cam = gl_ModelViewMatrixInverse * vec4(0.0,0.0,0.0,1.0);
  camVec = normalize(cam - vertex);
	refVec = vec3(cam) - vec3(vertex);
  camVector = camVec.xyz;

  // Figure out the upvector in model coordinates
  up = gl_ModelViewMatrixInverse * vec4(0.0, 1.0, 0.0, 1.0);
  up = normalize(up - cam);
  up = up - dot(up, camVec) * camVec;
  upVector = up.xyz;
  hint = dot(upVector,camVector);

  // Calculate rotation from (0,0,1) to cam
  // So this maps local coordinate center (0,0,1)
  // to world coordinate center
  rotationAxis = normalize(cross(vec3(0,0,1), camVec.xyz));
  rotationAngle = acos(dot(vec3(0,0,1), camVec.xyz));
  s1.xyz = sin(rotationAngle/2) * rotationAxis;
  s1.w = cos(rotationAngle/2);
  s1inv = quaternion_inverse(s1);

  // Rotate the local up-vector (0,1,0) to the world-coordinates
  upProj = quaternion_mult( s1, quaternion_mult(vec4(0,1,0,0), s1inv)).xyz;

  // The location of the up vector in world coordinates as given
  // by upProj probably not yet right. It needs to be <<up>>.
  // So calculate rotation from upProj to up.
  rotationAxis = normalize(cross(upProj.xyz, up.xyz));
  rotationAngle = acos(dot(upProj.xyz, up.xyz));
  s2.xyz = sin(rotationAngle/2) * rotationAxis;
  s2.w = cos(rotationAngle/2);

  // Combined rotation is given by s2 * s1
  s1 = quaternion_mult(s2, s1);

  // Normalize quaternion
  a = s1.w; b = s1.x; c = s1.y; d = s1.z;
  rotationMatrix = mat3( vec3(a*a + b*b - c*c - d*d, 2*b*c + 2*a*d, 2*b*d-2*a*c),
                           vec3(2*b*c-2*a*d, a*a - b*b + c*c - d*d, 2*c*d + 2*a*b),
						   vec3(2*b*d + 2*a*c, 2*c*d-2*a*b, a*a - b*b - c*c + d*d));

  // Determine the displacement (in the direction of the eye) to the surface
  displ = radius * normalize(camVec);

  // Transform the point on the surface to eye coordinates
  referenceVector = gl_ModelViewProjectionMatrix * (displ + vertex);

  // Calculate its (would be) fragment depth
  depth_top = (referenceVector.z / referenceVector.w) / 2.0 + 0.5;

  coord = (1.0 + halo_radius) * coord;

	/* Apply attenuation */
	distCam = (length(refVec) - PROTSTART) / (PROTEND - PROTSTART);
	attenuation = mix(1.0, 1.5 * (1.0 - distCam), sphere_attenuation);

  pos = gl_ModelViewMatrix * vertex;
  pos.x = pos.x + coord.s * attenuation * radius;
  pos.y = pos.y + coord.t * attenuation * radius;

  return pos;
}

void main() {
  vec4 coord;
  vec4 pos;
  float blend;
  
	illustrativeness = illustrativenessIn[0];
  structural_abstraction = structural_abs[0];
  if (structural_abstraction > 0.6)
    return;

	if (depth_cueing < 0.5) {
		sphere_attenuation = 2.0 * depth_cueing * attenuate;
		halo_radius = 0.0;
	}
	else {
		sphere_attenuation = 1.0 * attenuate;
		halo_radius = 0.4 * (2.0 * depth_cueing - 1.0);
	}

  calculate_abstraction();

	if (atom_rank[0] <= mol_abstraction) {
		if (depth_cueing < 0.5)
			ao_contrib = 2.0 * depth_cueing;
		else
			ao_contrib = 1.0;

		if (structural_abstraction < AO_ABS_LVL1) {
			blend = structural_abstraction / AO_ABS_LVL1;
			ao_fact = mix(texture2D(ambient_occlusion0, gl_TexCoordIn[0][0].pq).r,
										1.0, blend);
		}
		else {
			ao_fact = 1.0;
		}

		for(int i = 0; i < 4; ++i) {
			coord = coords[i];
			gl_FrontColor = gl_FrontColorIn[0];
			pastel_dark = pastel_darkIn[0];
			pastel_light = pastel_lightIn[0];
			//gl_FrontColor.a = mol_abstraction + atom_rank[0];

			pos = determine_position(gl_PositionIn[0], coord);
			gl_TexCoord[0] = coord;
			gl_TexCoord[0].pq = gl_TexCoordIn[0][0].pq;
			gl_Position = gl_ProjectionMatrix * pos;
			id = atom_id[0];
			type = atom_type[0];
			EmitVertex();
		}
	}
}

